# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import sys
from typing import List
from touca.cli._common import Operation, invalid_subcommand
from touca._options import (
    config_file_parse,
    find_home_path,
    find_profile_path,
)


class Config(Operation):
    name = "config"
    help = "Manage active configuration file"

    @classmethod
    def parser(cls, parser):
        parsers = parser.add_subparsers(dest="subcommand")
        parsers.add_parser(
            "home",
            description="Print path to active configuration file",
            help="Print path to active configuration file",
        )
        parsers.add_parser(
            "show",
            description="Print content of active configuration file",
            help="Print content of active configuration file",
        )
        parsers_1 = parsers.add_parser(
            "set",
            description="Set a value for a configuration option",
            help="Set a value for a configuration option",
        )
        parsers_1.add_argument(
            "key", nargs="+", help="option to be added to the config file"
        )
        parsers_2 = parsers.add_parser(
            "get",
            description="Get value of a configuration option",
            help="Get value of a configuration option",
        )
        parsers_2.add_argument("key", help="name of the option to print")
        parsers_3 = parsers.add_parser(
            "rm",
            description="Remove a configuration option",
            help="Remove a configuration option",
        )
        parsers_3.add_argument("key", help="name of the option to be removed")

    def __init__(self, options: dict):
        self.__options = options

    def _config_file_set(self, key: str, value: str, section="settings") -> None:
        from configparser import ConfigParser

        path = find_profile_path()
        path.parent.mkdir(parents=True, exist_ok=True)
        config = ConfigParser()
        if path.exists():
            config.read_string(path.read_text())
        if not config.has_section(section):
            config.add_section(section)
        config.set(section, key, value)
        with open(path, "wt") as file:
            config.write(file)

    def _command_home(self):
        print(find_home_path())
        return True

    def _command_show(self):
        path = find_profile_path()
        if path.exists():
            print(path.read_text().strip())
        return True

    def _command_get(self):
        key = self.__options.get("key")
        config = config_file_parse()
        if config and config.has_option("settings", key):
            print(config.get("settings", key))
            return True
        return False

    def _command_set(self):
        error = """
Argument \"{}\" has invalid format.

Please specify the value of the configuration option in the form of "key=value".
For example, to set Touca API Key in the configuration file you could write:

touca config set api-key=3c335732-bf44-4b28-9be8-f30c00e7960f
"""
        values: List[str] = self.__options.get("key")
        pairs = [x.split("=", maxsplit=1) for x in values]
        invalid = list(filter(lambda x: len(x) != 2, pairs))
        if invalid:
            print(str.format(error, invalid[0][0]), file=sys.stderr)
            return True
        for key, value in pairs:
            self._config_file_set(key, value)
        return True

    def _command_rm(self):
        key = self.__options.get("key")
        config_file_path = find_profile_path()
        config = config_file_parse()
        if config and config.has_option("settings", key):
            config.remove_option("settings", key)
        with open(config_file_path, "wt") as config_file:
            config.write(config_file)
        return True

    def run(self) -> bool:
        commands = {
            "home": self._command_home,
            "show": self._command_show,
            "get": self._command_get,
            "set": self._command_set,
            "rm": self._command_rm,
        }
        command = self.__options.get("subcommand")
        if not command:
            return invalid_subcommand(Config)
        if command in commands:
            return commands.get(command)()
        return False



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

from argparse import ArgumentParser
from pathlib import Path
import py7zr
from loguru import logger
from touca.cli._common import Operation


def _compress(srcDir, outputDir):
    dstFile = Path(outputDir, srcDir.name + ".7z")
    if dstFile.exists():
        logger.warning(f"compressed file {dstFile} already exists")
        return False
    logger.info(f"compressing {srcDir} into {dstFile}")
    try:
        with py7zr.SevenZipFile(dstFile, "w") as archive:
            archive.writeall(srcDir, arcname=srcDir.name)
    except py7zr.ArchiveError:
        logger.warning(f"failed to compress {srcDir}")
        return False
    return True


class Zip(Operation):
    name = "zip"
    help = "Compress binary archive files"

    def __init__(self, options: dict):
        self.__options = options

    @classmethod
    def parser(self, parser: ArgumentParser):
        parser.add_argument("src", help="directory with binary files")
        parser.add_argument("out", help="directory to store compressed files")

    def run(self):
        src = Path(self.__options.get("src")).expanduser().resolve()
        out = Path(self.__options.get("out")).expanduser().resolve()

        if not src.exists():
            logger.error(f"directory {src} does not exist")
            return False
        for src_dir in src.glob("*"):
            if not src_dir.is_dir():
                continue
            logger.debug(f"compressing {src_dir}")
            if not out.exists():
                out.mkdir(parents=True, exist_ok=True)
            if not _compress(src_dir, out):
                logger.error(f"failed to compress {src_dir}")
                return False
            logger.info(f"compressed {src_dir}")
        logger.info("compressed all sub-directories")
        return True



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

from touca._client import Client


class scoped_timer:
    def __init__(self, name):
        """
        Creates a timer object that captures its own lifetime and associates it
        with a given name.

        :param name: name to be associated with the captured runtime
        """
        self._name = name

    def __enter__(self):
        Client.instance().start_timer(self._name)

    def __exit__(self, exc_type, exc_value, traceback):
        Client.instance().stop_timer(self._name)



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

import certifi
import json
from touca.__init__ import __version__ as client_version
from typing import List
from urllib3.poolmanager import PoolManager
from urllib3.response import HTTPResponse


class Transport:
    """ """

    def __init__(self, options: dict):

        self._options = options
        self._token = None
        self._pool = PoolManager(cert_reqs="CERT_REQUIRED", ca_certs=certifi.where())
        self._handshake()

    def _handshake(self):
        response = self._send_request("GET", "/platform")
        if response.status != 200:
            raise RuntimeError("could not communicate with touca server")
        content = json.loads(response.data.decode("utf-8"))
        if not content["ready"]:
            raise RuntimeError("touca server is not ready")

    def _send_request(
        self, method, path, body=None, content_type="application/json"
    ) -> HTTPResponse:
        if body and content_type == "application/json":
            body = json.dumps(body).encode("utf-8")
        headers = {
            "Accept-Charset": "utf-8",
            "Accept": "application/json",
            "Content-Type": content_type,
            "User-Agent": f"touca-client-python/{client_version}",
        }
        if self._token:
            headers.update({"Authorization": f"Bearer {self._token}"})
        return self._pool.request(
            method=method,
            url=f"{self._options.get('api-url')}{path}",
            body=body,
            headers=headers,
        )

    def update_options(self, options: dict):
        if any(k in options for k in ["api-key", "api-url"]):
            self._token = None
            self._handshake()
        self._options.update(options)

    def authenticate(self):
        """
        :raises: ValueError if authentication fails
        """
        if self._token:
            return
        response = self._send_request(
            method="POST",
            path="/client/signin",
            body={"key": self._options.get("api-key")},
        )
        if response.status == 401:
            raise ValueError("Authentication failed: API Key Invalid")
        if response.status != 200:
            raise ValueError("Authentication failed: Invalid Response")
        body = json.loads(response.data.decode("utf-8"))
        self._token = body["token"]

    def get_testcases(self) -> List[str]:
        team = self._options.get("team")
        suite = self._options.get("suite")
        response = self._send_request(
            method="GET", path=f"/client/element/{team}/{suite}"
        )
        if response.status != 200:
            raise RuntimeError("Failed to obtain list of test cases")
        body = json.loads(response.data.decode("utf-8"))
        return [k["name"] for k in body]

    def post(self, content):
        response = self._send_request(
            method="POST",
            path=f"/client/submit",
            body=content,
            content_type="application/octet-stream",
        )
        if response.status == 204:
            return
        if response.status == 400 and "batch is sealed" in response.data.decode(
            "utf-8"
        ):
            reason = " This version is already submitted and sealed."
        raise RuntimeError(f"Failed to submit test results.{reason}")

    def seal(self):
        slugs = "/".join(self._options.get(k) for k in ["team", "suite", "version"])
        response = self._send_request(method="POST", path=f"/batch/{slugs}/seal2")
        if response.status != 204:
            raise RuntimeError("Failed to seal this version")



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

import math
from colorama import Style, Fore, Back, init

init()


class Printer:
    def print_warning(fmt: str, *args, **kwargs):
        print(f"{Fore.YELLOW}{fmt.format(*args, **kwargs)}{Fore.RESET}")

    def print_error(fmt: str, *args, **kwargs):
        import sys

        print(f"{Fore.RED}{fmt.format(*args, **kwargs)}{Fore.RESET}", file=sys.stderr)

    def print_app_header():
        print("\nTouca Test Framework")

    def print_app_footer():
        print("\n✨   Ran all test suites.\n")

    def __init__(self, options):
        self.options = options
        self.testcase_width = max(len(k) for k in options.get("testcases"))
        self.testcase_count = len(options.get("testcases"))

    def print_line(self, fmt: str, *args, **kwargs):
        msg = fmt.format(*args, **kwargs) if args or kwargs else fmt
        if self.options.get("colored-output"):
            print(msg)
            return

        import re

        line = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])").sub("", msg)
        print(line)

    def print_header(self):
        revision = "/".join([self.options.get(k) for k in ["suite", "version"]])
        self.print_line("\nSuite: {:s}\n", revision)

    def print_progress(self, timer, testcase, idx, status, errors=[]):
        states = {
            "pass": ("PASS", Back.GREEN),
            "skip": ("SKIP", Back.YELLOW),
            "fail": ("FAIL", Back.RED),
        }
        performance = (
            ""
            if status == "skip"
            else " {dim}({timer:d} ms){reset}".format(
                dim=Style.DIM,
                reset=Style.NORMAL,
                timer=timer.count(testcase),
            )
        )
        progress = " {number:>{width}d}{dim}.{reset}".format(
            dim=Style.DIM,
            reset=Style.NORMAL,
            number=idx + 1,
            count=self.testcase_count,
            width=int(math.log10(self.testcase_count)) + 1,
        )
        badge = "{bg_color} {text} {bg_reset}".format(
            bg_color=states.get(status)[1],
            bg_reset=Back.RESET,
            text=states.get(status)[0],
        )
        self.print_line(
            "{progress} {badge}  {testcase:<{testcase_width}s}{performance}",
            badge=badge,
            progress=progress,
            testcase=testcase,
            testcase_width=self.testcase_width + 3,
            performance=performance,
        )
        if errors:
            self.print_line("\n   {}Exception Raised:{}", Style.DIM, Style.NORMAL)
            self.print_line("\n".join(f"      - {error}\n" for error in errors))

    def print_footer(self, stats, timer):
        states = [
            ("pass", "passed", Fore.GREEN),
            ("skip", "skipped", Fore.YELLOW),
            ("fail", "failed", Fore.RED),
        ]
        messages = []
        for state in states:
            if not stats.count(state[0]):
                continue
            messages.append(f"{state[2]}{stats.count(state[0])} {state[1]}{Fore.RESET}")
        messages.append(f"{self.testcase_count} total")
        self.print_line(
            "\n{:s} {:s}",
            "Tests:".ljust(int(math.log10(self.testcase_count)) + 11),
            ", ".join(messages),
        )
        self.print_line(
            "{:s} {:.2f} s",
            "Time:".ljust(int(math.log10(self.testcase_count)) + 11),
            timer.count("__workflow__") / 1000,
        )



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

from argparse import ArgumentParser
from pathlib import Path
import py7zr
from loguru import logger
from touca.cli._common import Operation


def _extract(src, dst):
    try:
        logger.info(f"extracting file {src} into {dst}")
        with py7zr.SevenZipFile(src, "r") as archive:
            archive.extractall(path=dst)
    except Exception:
        logger.warning(f"failed to extract {src}")
        return False
    return True


class Unzip(Operation):
    name = "unzip"
    help = "Extract compressed binary archive files"

    def __init__(self, options: dict):
        self.__options = options

    @classmethod
    def parser(self, parser: ArgumentParser):
        parser.add_argument("src", help="directory with compressed files")
        parser.add_argument("out", help="directory to extract binary files into")

    def run(self):
        src = Path(self.__options.get("src")).expanduser().resolve()
        out = Path(self.__options.get("out")).expanduser().resolve()

        if not src.exists():
            logger.error(f"directory {src} does not exist")
            return False
        for src_dir in src.glob("*"):
            if not py7zr.is_7zfile(src_dir):
                logger.debug(f"{src_dir} is not an archive file")
                continue
            dstDir = out.joinpath(src_dir.stem)
            if dstDir.exists():
                logger.debug(f"unzipped directory already exists: {dstDir}")
                continue
            out.mkdir(parents=True, exist_ok=True)
            if not _extract(src_dir, out):
                return False
            logger.info(f"extracted {src_dir}")
        logger.info("extracted all archives")
        return True



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

from sys import stderr, stdout
from pathlib import Path
from argparse import ArgumentParser
from loguru import logger
from touca.cli._common import Operation


def _merge(touca_cli: Path, dir_src: Path, dir_dst: Path):
    from subprocess import Popen

    if not dir_src.exists():
        logger.error(f"expected directory {dir_src} to exist")
        return False
    dir_dst.mkdir(parents=True, exist_ok=True)
    logger.info(f"merging result directory {dir_src} into {dir_dst}")
    cmd = [touca_cli, "merge", f"--src={dir_src}", f"--out={dir_dst}"]
    proc = Popen(cmd, universal_newlines=True, stdout=stdout, stderr=stderr)
    exit_status = proc.wait()
    if 0 != exit_status:
        logger.warning(f"failed to merge {dir_src}")
        if exit_status is not None:
            logger.warning(f"touca_cli returned code {exit_status}")
        return False
    return True


class Merge(Operation):
    name = "merge"
    help = "Merge binary archive files"

    def __init__(self, options: dict):
        self.__options = options

    @classmethod
    def parser(self, parser: ArgumentParser):
        parser.add_argument(
            "--src",
            help="path to directory with original Touca archives directories",
            required=True,
        )
        parser.add_argument(
            "--out",
            help="path to directory where the merged archives should be created",
            required=True,
        )
        parser.add_argument(
            "--cli",
            help='path to "touca_cli" C++ executable',
            required=True,
        )

    def run(self):
        src = Path(self.__options.get("src")).expanduser().resolve()
        out = Path(self.__options.get("out")).expanduser().resolve()
        cli = Path(self.__options.get("cli")).expanduser().resolve()

        if not src.exists():
            logger.error(f"directory {src} does not exist")
            return False
        if not out.exists():
            out.mkdir(parents=True, exist_ok=True)

        for dir_src in src.glob("*"):
            if not dir_src.is_dir():
                continue
            if dir_src.name.endswith("-merged"):
                continue
            dir_dst = out.joinpath(dir_src.name + "-merged")
            if dir_dst.exists():
                continue
            logger.info(f"merging {dir_src}")
            if not _merge(cli, dir_src, dir_dst):
                logger.error(f"failed to merge {dir_src}")
                return False
            logger.info(f"merged {dir_src}")
        logger.info("merged all result directories")
        return True



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

from argparse import ArgumentParser
import sys

from loguru import logger
from touca import __version__
from touca._options import find_home_path
from touca._printer import Printer
from touca.cli._config import Config
from touca.cli._execute import Execute
from touca.cli._merge import Merge
from touca.cli._profile import Profile
from touca.cli._post import Post
from touca.cli._run import Run
from touca.cli._solve import Solve
from touca.cli._unzip import Unzip
from touca.cli._update import Update
from touca.cli._zip import Zip


def _find_latest_pypi_version():
    from json import loads
    from urllib.request import urlopen

    with urlopen("https://pypi.org/pypi/touca/json") as response:
        data = loads(response.read())
        return data["info"]["version"]


def _warn_outdated_version():
    from packaging import version

    latest_version = _find_latest_pypi_version()
    if version.parse(latest_version) <= version.parse(__version__):
        return
    fmt = (
        "You are using touca version {}; however, version {} is available."
        + "\nConsider upgrading by running 'pip install --upgrade touca'."
    )
    Printer.print_warning(fmt, __version__, latest_version)


def main(args=None):
    operations = [Config, Merge, Post, Profile, Run, Solve, Execute, Unzip, Update, Zip]
    parser = ArgumentParser(
        prog="touca",
        add_help=False,
        description="Work seamlessly with Touca from the command line.",
        epilog="See https://touca.io/docs for more information.",
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"%(prog)s v{__version__}"
    )
    parsers = parser.add_subparsers(dest="command")
    for operation in operations:
        subparser = parsers.add_parser(
            name=operation.name,
            prog=f"touca {operation.name}",
            description=operation.help,
            help=operation.help,
            add_help=True,
        )
        operation.parser(subparser)
    parsed, remaining = parser.parse_known_args(sys.argv[1:] if args is None else args)
    options = vars(parsed)

    command = next((x for x in operations if x.name == options.get("command")), None)
    if not command and any(arg in remaining for arg in ["-h", "--help"]):
        parser.print_help()
        return False
    operation = command(options) if command else Execute(options)

    home_dir = find_home_path()
    home_dir.mkdir(parents=True, exist_ok=True)

    logger.remove()
    logger.add(
        sys.stdout,
        level="DEBUG",
        colorize=True,
        format="<green>{time:HH:mm:ss!UTC}</green> | <cyan>{level: <7}</cyan> | <lvl>{message}</lvl>",
    )
    logger.add(
        home_dir.joinpath("logs", "runner_{time:YYMMDD!UTC}.log"),
        level="DEBUG",
        rotation="1 day",
        compression="zip",
    )

    if not operation.run():
        return True

    _warn_outdated_version()
    return False


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

from json import dumps
from abc import ABC, abstractmethod
from collections.abc import Iterable
from flatbuffers import Builder
from typing import Any, Callable, Dict, Type
import touca._schema as schema


class ToucaType(ABC):
    @abstractmethod
    def json(self):
        pass

    @abstractmethod
    def serialize(self, builder: Builder):
        pass


class BoolType(ToucaType):
    def __init__(self, value: bool):
        """
        Converts a given value of primitive type ``bool`` to a BoolType with
        known serialization into JSON and Binary.
        """
        self._value = value

    def json(self):
        return self._value

    def serialize(self, builder: Builder):
        schema.BoolStart(builder)
        schema.BoolAddValue(builder, value=self._value)
        value = schema.BoolEnd(builder)
        schema.TypeWrapperStart(builder)
        schema.TypeWrapperAddValue(builder, value)
        schema.TypeWrapperAddValueType(builder, schema.Type.Bool)
        return schema.TypeWrapperEnd(builder)


class DecimalType(ToucaType):
    def __init__(self, value: float):

        self._value = value

    def json(self):
        return self._value

    def serialize(self, builder: Builder):
        schema.DoubleStart(builder)
        schema.DoubleAddValue(builder, self._value)
        value = schema.DoubleEnd(builder)
        schema.TypeWrapperStart(builder)
        schema.TypeWrapperAddValue(builder, value)
        schema.TypeWrapperAddValueType(builder, schema.Type.Double)
        return schema.TypeWrapperEnd(builder)


class IntegerType(ToucaType):
    def __init__(self, value: int):

        self._value = value

    def json(self):
        return self._value

    def serialize(self, builder: Builder):
        schema.IntStart(builder)
        schema.IntAddValue(builder, self._value)
        value = schema.IntEnd(builder)
        schema.TypeWrapperStart(builder)
        schema.TypeWrapperAddValue(builder, value)
        schema.TypeWrapperAddValueType(builder, schema.Type.Int)
        return schema.TypeWrapperEnd(builder)


class StringType(ToucaType):
    def __init__(self, value: str):
        self._value = value

    def json(self):
        return self._value

    def serialize(self, builder: Builder):
        content = Builder.CreateString(builder, self._value)
        schema.StringStart(builder)
        schema.StringAddValue(builder, content)
        value = schema.StringEnd(builder)
        schema.TypeWrapperStart(builder)
        schema.TypeWrapperAddValue(builder, value)
        schema.TypeWrapperAddValueType(builder, schema.Type.String)
        return schema.TypeWrapperEnd(builder)


class VectorType(ToucaType):
    def __init__(self):
        self._values = []

    def add(self, value: ToucaType):
        self._values.append(value)

    def json(self):
        return dumps([v.json() for v in self._values])

    def serialize(self, builder: Builder):
        items = [v.serialize(builder) for v in self._values]
        schema.ArrayStartValuesVector(builder, len(items))
        for item in items:
            builder.PrependUOffsetTRelative(item)
        values = builder.EndVector()
        schema.ArrayStart(builder)
        schema.ArrayAddValues(builder, values)
        value = schema.ArrayEnd(builder)
        schema.TypeWrapperStart(builder)
        schema.TypeWrapperAddValue(builder, value)
        schema.TypeWrapperAddValueType(builder, schema.Type.Array)
        return schema.TypeWrapperEnd(builder)


class ObjectType(ToucaType):
    def __init__(self, key: str):
        self._name = key
        self._values = {}

    def add(self, key: str, value: ToucaType):
        self._values[key] = value

    def json(self):
        return dumps({k: v.json() for k, v in self._values.items()})

    def serialize(self, builder: Builder):
        key = builder.CreateString(self._name)
        members = []
        for k, v in self._values.items():
            memberkey = builder.CreateString(k)
            membervalue = v.serialize(builder)
            schema.ObjectMemberStart(builder)
            schema.ObjectMemberAddName(builder, memberkey)
            schema.ObjectMemberAddValue(builder, membervalue)
            members.append(schema.ObjectMemberEnd(builder))
        schema.ObjectStartValuesVector(builder, len(members))
        for item in reversed(members):
            builder.PrependUOffsetTRelative(item)
        values = builder.EndVector()
        schema.ObjectStart(builder)
        schema.ObjectAddKey(builder, key)
        schema.ObjectAddValues(builder, values)
        value = schema.ObjectEnd(builder)
        schema.TypeWrapperStart(builder)
        schema.TypeWrapperAddValue(builder, value)
        schema.TypeWrapperAddValueType(builder, schema.Type.Object)
        return schema.TypeWrapperEnd(builder)


class TypeHandler:
    def __init__(self):
        from datetime import date

        self._primitives: Dict[Type, Callable[[Any], ToucaType]] = {
            bool: BoolType,
            float: DecimalType,
            int: IntegerType,
            str: StringType,
        }
        self._types = {date: (lambda x: dict(year=x.year, month=x.month, day=x.day))}

    def transform(self, value: Any):
        if type(value) in self._primitives:
            return self._primitives.get(type(value))(value)
        if type(value) in self._types:
            return self.transform(self._types.get(type(value))(value))
        if isinstance(value, dict):
            obj = ObjectType(value.__class__.__name__)
            for k, v in value.items():
                obj.add(k, self.transform(v))
            return obj
        if isinstance(value, Iterable):
            vec = VectorType()
            for v in value:
                vec.add(self.transform(v))
            return vec
        return self._objectify(value)

    def add_serializer(self, datatype: Type, serializer: Callable[[Any], Dict]):
        self._types[datatype] = serializer

    def _objectify(self, value: Any):
        obj = ObjectType(value.__class__.__name__)
        if hasattr(value, "__slots__"):
            if isinstance(value, Iterable):
                for k in value.__slots__:
                    obj.add(k, self.transform(getattr(value, k)))
        elif hasattr(value, "__dict__"):
            for k, v in value.__dict__.items():
                obj.add(k, self.transform(v))
        else:
            for k in dir(value):
                v = getattr(value, k)
                if k.startswith("_") or callable(v) or type(value) == type(v):
                    continue
                obj.add(k, self.transform(v))
        return obj



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

from subprocess import Popen
from sys import stderr, stdout
from pathlib import Path
from argparse import ArgumentParser
from distutils.version import LooseVersion
from loguru import logger
from touca.cli._common import Operation


def _update(cli, srcDir: Path, outDir: Path, teamslug, testsuite):
    print(srcDir)
    srcDir = srcDir.with_name(srcDir.name + "-merged")
    if not srcDir.exists():
        logger.error(f"expected directory {srcDir} to exist")
        return False
    outDir.mkdir(parents=True, exist_ok=True)
    logger.info(f"updating result directory {srcDir}")
    cmd = [
        cli,
        "update",
        f"--src={srcDir}",
        f"--out={outDir}",
        f"--team={teamslug}",
        f"--suite={testsuite}",
    ]
    proc = Popen(cmd, universal_newlines=True, stdout=stdout, stderr=stderr)
    exit_status = proc.wait()
    if 0 != exit_status:
        logger.warning(f"failed to update {srcDir}")
        if exit_status is not None:
            logger.warning(f"program Touca Utils returned code {exit_status}")
        return False
    logger.info(f"updated {srcDir}")
    return True


class Update(Operation):
    name = "update"
    help = "Update metdata of binary archive files"

    def __init__(self, options: dict):
        self.__options = options

    @classmethod
    def parser(self, parser: ArgumentParser):
        parser.add_argument(
            "--src",
            help="path to directory with original Touca binary archives",
            required=True,
        )
        parser.add_argument(
            "--out",
            help="path to directory with updated Touca binary archives",
            required=True,
        )
        parser.add_argument(
            "--cli",
            help='path to "touca_cli" C++ executable',
            required=True,
        )
        parser.add_argument("--team", help="new value for the team slug")
        parser.add_argument("--suite", help="new value for the suite slug")

    def run(self):
        src = Path(self.__options.get("src")).expanduser().resolve()
        out = Path(self.__options.get("out")).expanduser().resolve()
        cli = Path(self.__options.get("cli")).expanduser().resolve()
        teamslug = self.__options.get("team")
        testsuite = self.__options.get("suite")

        if not src.exists():
            logger.error(f"directory {src} does not exist")
            return False
        batchNames = []
        for batchDir in src.glob("*"):
            print(batchDir)
            if not batchDir.is_dir():
                continue
            if not batchDir.name.endswith("-merged"):
                continue
            batchNames.append(batchDir.name[:-7])

        if not batchNames:
            logger.info(f"found no valid result directory to update")
            return True

        # sort list of versions lexicographically
        batchNames.sort(key=LooseVersion)

        logger.info(f"updating batches one by one")
        for batchName in batchNames:
            batchDir = src.joinpath(batchName)
            logger.info(f"updating {batchDir}")
            updated = _update(
                cli, batchDir, out.joinpath(batchName), teamslug, testsuite
            )
            if not updated:
                logger.error(f"failed to update {batchDir}")
                return False
        logger.info("updated all result directories")
        return True



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

from argparse import ArgumentParser
from pathlib import Path
import sys
from touca._options import config_file_parse
from touca._runner import prepare_parser, run_workflows, Workflow
from touca.cli._common import Operation


def is_test_module(module: str):
    with open(module, "rt") as file:
        return "@touca.Workflow" in file.read()


def find_test_modules(testdir: str):
    return [p for p in Path(testdir).glob("**/*.py") if is_test_module(p)]


def extract_workflows(modules: list):
    import sys
    import importlib
    import inspect

    for module in modules:
        relpath = Path(module).relative_to(Path.cwd())
        syspath = Path(relpath.parent).absolute()
        sys.path.append(f"{syspath}/")
        mod = importlib.import_module(relpath.stem)
        for (name, member) in inspect.getmembers(mod):
            if isinstance(member, Workflow):
                yield name, member
        sys.path.remove(f"{syspath}/")


class Execute(Operation):
    name = "test"
    help = "Execute available regression tests"

    @classmethod
    def parser(cls, parser: ArgumentParser):
        parser.add_argument(
            "--testdir",
            default=[Path.cwd()],
            nargs=1,
            help="path to regression tests directory",
        )
        prepare_parser(parser)

    def __init__(self, options: dict):
        self.__options = {k: v for k, v in options.items() if v is not None}

    def _find_arguments(self):
        args = self.__options
        flags = [
            "save-as-binary",
            "save-as-json",
            "offline",
            "overwrite",
            "colored-output",
        ]
        config_content = config_file_parse()
        if config_content:
            for key in config_content.options("settings"):
                func = config_content.getboolean if key in flags else config_content.get
                args.update({key: func("settings", key)})
        return args

    def run(self):
        dir_test = Path(self.__options.get("testdir")[0]).resolve()
        args = self._find_arguments()
        modules = find_test_modules(dir_test)
        workflows = list(extract_workflows(modules))
        try:
            run_workflows(args, workflows)
        except Exception as err:
            print(f"test failed: {err}", file=sys.stderr)
            return False
        return True



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

from argparse import ArgumentParser

from touca.cli._common import Operation


class Solve(Operation):
    name = "solve"
    help = "Solve every problem (April 1st, 2022)"

    def __init__(self, options: dict):
        pass

    @classmethod
    def parser(self, parser: ArgumentParser):
        parser.add_argument("args", nargs="+", help="any problem")

    def run(self):
        print("Solved!")
        return True



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

from pathlib import Path
from configparser import ConfigParser


def _apply_legacy_config_file(incoming: dict) -> None:
    from json import loads

    if "file" not in incoming:
        return
    path = Path(incoming.get("file"))
    if not path.is_file():
        raise ValueError("configuration file not found")
    content = path.read_text()
    try:
        parsed = loads(content)
    except ValueError:
        raise ValueError("configuration file has unexpected format")
    if "touca" not in parsed:
        raise ValueError('configuration file is missing JSON field: "touca"')
    for k in parsed["touca"]:
        if k not in incoming:
            incoming[k] = parsed["touca"][k]


def _apply_config_file(incoming: dict) -> None:
    config = config_file_parse()
    if not config or not config.has_section("settings"):
        return
    for key in config.options("settings"):
        if key not in incoming:
            incoming[key] = config.get("settings", key)


def _apply_arguments(existing, incoming: dict) -> None:
    for params, validate in [
        (
            ["team", "suite", "version", "api-key", "api-url"],
            lambda x: isinstance(x, str),
        ),
        (["offline", "concurrency"], lambda x: isinstance(x, bool)),
    ]:
        for param in params:
            if param not in incoming:
                continue
            value = incoming.get(param)
            if not validate(value):
                raise ValueError(f"parameter {param} has unexpected type")
            existing[param] = value


def _apply_environment_variables(existing) -> None:
    from os import environ

    for env, opt in [
        ("TOUCA_API_KEY", "api-key"),
        ("TOUCA_API_URL", "api-url"),
        ("TOUCA_TEST_VERSION", "version"),
    ]:
        if environ.get(env):
            existing[opt] = environ.get(env)


def _reformat_parameters(existing: dict) -> None:
    from urllib.parse import urlparse

    existing.setdefault("concurrency", True)

    api_url = existing.get("api-url")
    if not api_url:
        # existing["api-url"] = "https://api.touca.io"
        return
    url = urlparse(api_url)
    urlpath = [k.strip("/") for k in url.path.split("/@/")]
    existing["api-url"] = f"{url.scheme}://{url.netloc}/{urlpath[0]}".rstrip("/")

    if len(urlpath) == 1:
        return

    slugs = [k for k in urlpath[1].split("/") if k]
    for k, v in list(zip(["team", "suite", "version"], slugs)):
        existing[k] = v


def _validate_options(existing: dict):
    expected_keys = ["team", "suite", "version"]
    has_handshake = not existing.get("offline")
    if has_handshake and any(x in existing for x in ["api-key", "api-url"]):
        expected_keys.extend(["api-key", "api-url"])
    key_status = {k: k in existing for k in expected_keys}
    if any(key_status.values()) and not all(key_status.values()):
        keys = list(filter(lambda x: not key_status[x], key_status))
        raise ValueError(f"missing required option(s) {','.join(keys)}")


def update_options(existing: dict, incoming: dict):
    _apply_legacy_config_file(incoming)
    _apply_config_file(incoming)
    _apply_arguments(existing, incoming)
    _apply_environment_variables(existing)
    _reformat_parameters(existing)
    _validate_options(existing)


def find_home_path():
    path = Path.cwd().joinpath(".touca")
    return path if path.exists() else Path.home().joinpath(".touca")


def find_profile_path():
    home_path = find_home_path()
    settings_path = Path(home_path, "settings")
    name = "default"
    if settings_path.exists():
        config = ConfigParser()
        config.read_string(settings_path.read_text())
        name = config.get("settings", "profile")
    return home_path.joinpath("profiles", name)


def config_file_parse() -> ConfigParser:
    path = find_profile_path()
    if path.exists():
        config = ConfigParser()
        config.read_string(path.read_text())
        return config



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

from touca._types import IntegerType, VectorType, ToucaType
from datetime import datetime, timedelta
from enum import Enum
from typing import Dict, Tuple


class ResultCategory(Enum):
    """ """

    Check = 1
    Assert = 2


class ResultEntry:
    """
    Wrapper around a given ``ToucaType`` value that includes the category
    it should belong to.

    We are intentionally not using ``@dataclass`` to ensure the core library
    has no dependency on ``dataclasses`` module. This may change in the future.
    """

    def __init__(self, typ: ResultCategory, val: ToucaType):
        """
        Creates an entry given its value and the category it should belong to.

        :param typ: type of the entry
        :param val: value of the entry
        """
        self.typ = typ
        self.val = val


class Case:
    """ """

    def __init__(self, **kwargs):

        self._meta = kwargs
        self._results: Dict[str, ResultEntry] = dict()
        self._tics: Dict[str, datetime] = dict()
        self._tocs: Dict[str, datetime] = dict()

    def check(self, key: str, value: ToucaType):
        """
        Logs a given value as a test result for the declared test case
        and associates it with the specified key.

        :param key: name to be associated with the logged test result
        :param value: value to be logged as a test result
        """
        self._results[key] = ResultEntry(typ=ResultCategory.Check, val=value)

    def assume(self, key: str, value: ToucaType):
        """
        Logs a given value as an assertion for the declared test case
        and associates it with the specified key.

        :param key: name to be associated with the logged test result
        :param value: value to be logged as a test result
        """
        self._results[key] = ResultEntry(typ=ResultCategory.Assert, val=value)

    def add_array_element(self, key: str, value: ToucaType):
        """
        Adds a given value to a list of results for the declared
        test case which is associated with the specified key.

        Could be considered as a helper utility function.
        This method is particularly helpful to log a list of items as they
        are found:

        .. code-block:: python

            for number in numbers:
                if is_prime(number):
                    touca.add_array_element("prime numbers", number)
                    touca.add_hit_count("number of primes")

        This pattern can be considered as a syntactic sugar for the following
        alternative:

        .. code-block:: python

            primes = []
            for number in numbers:
                if is_prime(number):
                    primes.append(number)
            if primes:
                touca.check("prime numbers", primes)
                touca.check("number of primes", len(primes))

        The items added to the list are not required to be of the same type.
        The following code is acceptable:

        .. code-block:: python

            touca.check("prime numbers", 42)
            touca.check("prime numbers", "forty three")

        :raises RuntimeError:
            if specified key is already associated with
            a test result which was not iterable

        :param key: name to be associated with the logged test result
        :param value: element to be appended to the array
        :see also: :py:meth:`~check`
        """
        if key not in self._results:
            self._results[key] = ResultEntry(typ=ResultCategory.Check, val=VectorType())
        vec = self._results.get(key)
        if vec.typ is not ResultCategory.Check or not isinstance(vec.val, VectorType):
            raise RuntimeError("specified key has a different type")
        vec.val.add(value)

    def add_hit_count(self, key: str):
        """
        Increments value of key every time it is executed.
        creates the key with initial value of one if it does not exist.

        Could be considered as a helper utility function.
        This method is particularly helpful to track variables whose values
        are determined in loops with indeterminate execution cycles:

        .. code-block:: python

            for number in numbers:
                if is_prime(number):
                    touca.add_array_element("prime numbers", number)
                    touca.add_hit_count("number of primes")

        This pattern can be considered as a syntactic sugar for the following
        alternative:

        .. code-block:: python

            primes = []
            for number in numbers:
                if is_prime(number):
                    primes.append(number)
            if primes:
                touca.check("prime numbers", primes)
                touca.check("number of primes", len(primes))

        :raises RuntimeError:
            if specified key is already associated with
            a test result which was not an integer

        :param key: name to be associated with the logged test result
        :see also: :py:meth:`~check`
        """
        if key not in self._results:
            self._results[key] = ResultEntry(
                typ=ResultCategory.Check, val=IntegerType(1)
            )
            return
        value = self._results.get(key)
        if value.typ is not ResultCategory.Check or not isinstance(
            value.val, IntegerType
        ):
            raise RuntimeError("specified key has a different type")
        value.val._value += 1

    def add_metric(self, key: str, milliseconds: int):
        """
        Adds an already obtained measurements to the list of captured
        performance benchmarks.

        Useful for logging a metric that is measured without using this SDK.

        :param key: name to be associated with this performance benchmark
        :param milliseconds: duration of this measurement in milliseconds
        """
        value = datetime.now()
        self._tics[key] = value
        self._tocs[key] = value + timedelta(microseconds=milliseconds * 1000)

    def start_timer(self, key: str):
        """
        Starts timing an event with the specified name.

        Measurement of the event is only complete when function
        :py:meth:`~stop_timer` is later called for the specified name.

        :param key: name to be associated with the performance metric
        """
        self._tics[key] = datetime.now()

    def stop_timer(self, key: str):
        """
        Stops timing an event with the specified name.

        Expects function :py:meth:`~start_timer` to have been called previously
        with the specified name.

        :param key: name to be associated with the performance metric
        """
        if key in self._tics:
            self._tocs[key] = datetime.now()

    def _metrics(self) -> Tuple[str, ToucaType]:
        for key, tic in self._tics.items():
            if key not in self._tocs:
                continue
            diff = (self._tocs.get(key) - tic).microseconds / 1000
            yield key, IntegerType(int(diff))

    def _metadata(self) -> Dict[str, str]:
        return {
            "teamslug": self._meta.get("team") or "unknown",
            "testsuite": self._meta.get("suite") or "unknown",
            "version": self._meta.get("version") or "unknown",
            "testcase": self._meta.get("name") or "unknown",
            "builtAt": datetime.now().isoformat(),
        }

    def json(self):
        return {
            "metadata": self._metadata(),
            "results": [
                {"key": k, "value": v.val.json()}
                for k, v in self._results.items()
                if v.typ is ResultCategory.Check
            ],
            "assertions": [
                {"key": k, "value": v.val.json()}
                for k, v in self._results.items()
                if v.typ is ResultCategory.Assert
            ],
            "metrics": [{"key": k, "value": v.json()} for k, v in self._metrics()],
        }

    def serialize(self) -> bytearray:
        from flatbuffers import Builder
        import touca._schema as schema

        dicts = {
            ResultCategory.Check: schema.ResultType.Check,
            ResultCategory.Assert: schema.ResultType.Assert,
        }
        builder = Builder(1024)

        metadata = {k: builder.CreateString(v) for k, v in self._metadata().items()}
        schema.MetadataStart(builder)
        schema.MetadataAddTeamslug(builder, metadata.get("teamslug"))
        schema.MetadataAddTestsuite(builder, metadata.get("testsuite"))
        schema.MetadataAddVersion(builder, metadata.get("version"))
        schema.MetadataAddTestcase(builder, metadata.get("testcase"))
        schema.MetadataAddBuiltAt(builder, metadata.get("builtAt"))
        fbs_metadata = schema.MetadataEnd(builder)

        result_entries = []
        for k, v in self._results.items():
            fbs_key = Builder.CreateString(builder, k)
            fbs_value = v.val.serialize(builder)
            schema.ResultStart(builder)
            schema.ResultAddKey(builder, fbs_key)
            schema.ResultAddValue(builder, fbs_value)
            schema.ResultAddTyp(builder, dicts.get(v.typ))
            result_entries.append(schema.ResultEnd(builder))

        schema.ResultsStartEntriesVector(builder, len(result_entries))
        for item in reversed(result_entries):
            builder.PrependUOffsetTRelative(item)
        fbs_result_entries = builder.EndVector()

        schema.ResultsStart(builder)
        schema.ResultsAddEntries(builder, fbs_result_entries)
        fbs_results = schema.ResultsEnd(builder)

        metric_entries = []
        for k, v in self._metrics():
            fbs_key = Builder.CreateString(builder, k)
            fbs_value = v.serialize(builder)
            schema.MetricStart(builder)
            schema.MetricAddKey(builder, fbs_key)
            schema.MetricAddValue(builder, fbs_value)
            metric_entries.append(schema.MetricEnd(builder))

        schema.MetricsStartEntriesVector(builder, len(metric_entries))
        for item in reversed(metric_entries):
            builder.PrependUOffsetTRelative(item)
        fbs_metric_entries = builder.EndVector()

        schema.MetricsStart(builder)
        schema.MetricsAddEntries(builder, fbs_metric_entries)
        fbs_metrics = schema.MetricsEnd(builder)

        schema.MessageStart(builder)
        schema.MessageAddMetadata(builder, fbs_metadata)
        schema.MessageAddResults(builder, fbs_results)
        schema.MessageAddMetrics(builder, fbs_metrics)
        fbs_message = schema.MessageEnd(builder)

        builder.Finish(fbs_message)
        return builder.Output()



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

from configparser import ConfigParser
from pathlib import Path
from argparse import ArgumentParser
import sys
from typing import List
from touca.cli._common import Operation, invalid_subcommand
from touca._options import find_home_path, find_profile_path


class Profile(Operation):
    name = "profile"
    help = "Create and manage configuration files"

    @classmethod
    def parser(cls, parser: ArgumentParser):
        parsers = parser.add_subparsers(dest="subcommand")
        parsers.add_parser(
            "list",
            description="List available profiles",
            help="List available profiles",
        )
        parsers_1 = parsers.add_parser(
            "set", description="Change active profile", help="Change active profile"
        )
        parsers_1.add_argument("name", help="name of the profile")
        parsers_2 = parsers.add_parser(
            "delete",
            description="Delete profile with specified name",
            help="Delete profile with specified name",
        )
        parsers_2.add_argument("name", help="name of the profile")
        parsers_3 = parsers.add_parser(
            "copy",
            description="Copy content of a given profile to a new or existing profile",
            help="Copy content of a profile to a new or existing profile",
        )
        parsers_3.add_argument("src", help="name of the profile to copy from")
        parsers_3.add_argument("dst", help="name of the new profile")

    def __init__(self, options: dict):
        self.__options = options

    def _make_profile(self, profile: Path):
        if profile.exists():
            return
        profile.parent.mkdir(parents=True, exist_ok=True)
        config = ConfigParser()
        config.add_section("settings")
        with open(profile, "wt") as config_file:
            config.write(config_file)

    def _update_profile_in_settings_file(self, profile_name: str):
        settings_path = Path(find_home_path(), "settings")
        config = ConfigParser()
        if settings_path.exists():
            config.read_string(settings_path.read_text())
        if not config.has_section("settings"):
            config.add_section("settings")
        config.set("settings", "profile", profile_name)
        with open(settings_path, "wt") as settings_file:
            config.write(settings_file)

    def _command_list(self):
        profiles_dir = Path(find_home_path(), "profiles")
        profile_names = [p.name for p in profiles_dir.glob("*") if p.is_file()]
        if "default" not in profile_names:
            self._make_profile(profiles_dir.joinpath("default"))
            profile_names.append("default")
        for name in profile_names:
            print(name)
        return True

    def _command_set(self):
        profile_name = self.__options.get("name")
        profile_path = Path(find_home_path(), "profiles", profile_name)
        self._make_profile(profile_path)
        self._update_profile_in_settings_file(profile_name)
        return True

    def _command_delete(self):
        profile_name = self.__options.get("name")
        profile_path = Path(find_home_path(), "profiles", profile_name)
        if profile_name == "default":
            print("refusing to remove default configuration file", file=sys.stderr)
            return False
        if not profile_path.exists():
            print("profile does not exist", file=sys.stderr)
            return False
        if profile_path == find_profile_path():
            self._update_profile_in_settings_file("default")
        profile_path.unlink()
        return True

    def _command_copy(self):
        from shutil import copyfile

        profiles_dir = Path(find_home_path(), "profiles")
        src = self.__options.get("src")
        dst = self.__options.get("dst")
        profile_path = Path(profiles_dir, src)
        if not profile_path.exists():
            print(f'profile "{src}" does not exist', file=sys.stderr)
            return False
        copyfile(profile_path, Path(profiles_dir, dst))
        return True

    def run(self):
        commands = {
            "list": self._command_list,
            "set": self._command_set,
            "delete": self._command_delete,
            "copy": self._command_copy,
        }
        command = self.__options.get("subcommand")
        if not command:
            return invalid_subcommand(Profile)
        if command in commands:
            return commands.get(command)()
        return False



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

"""
Touca Test Framework for Python is designed to make writing regression test
workflows easy and straightforward. The test framework abstracts away many
of the common expected features such as logging, error handling and reporting
test progress.

The following example demonstrates how to use this framework::

    import touca
    from code_under_test import find_student, calculate_gpa

    @touca.Workflow
    def test_students(testcase: str):
        student = find_student(testcase)
        touca.assume("username", student.username)
        touca.check("fullname", student.fullname)
        touca.check("birth_date", student.dob)
        touca.check("gpa", calculate_gpa(student.courses))

    if __name__ == "__main__":
        touca.run()

It is uncommon to run multiple regression test workflows as part of a single
suite. However, the pattern above allows introducing multiple workflows by
defining functions with ``@touca.Workflow`` decorators.
"""

import os
import shutil
import sys
import textwrap
from argparse import Action, ArgumentParser
from datetime import datetime, timedelta
from enum import IntEnum
from typing import Any, Dict, List
from touca._client import Client
from touca._printer import Printer


def prepare_parser(parser: ArgumentParser):
    class ExtendAction(Action):
        def __call__(self, parser, namespace, values, option_string=None):
            items = getattr(namespace, self.dest) or []
            items.extend(values)
            setattr(namespace, self.dest, items)

    parser.register("action", "extend", ExtendAction)
    parser.add_argument("--api-key", help="Touca API Key", dest="api-key")
    parser.add_argument(
        "--api-url",
        help="Touca API URL",
        dest="api-url",
        default="https://api.touca.io",
    )
    parser.add_argument(
        "--revision",
        help="Version of the code under test",
        dest="version",
    )
    parser.add_argument(
        "--suite",
        help="Slug of suite to which test results belong",
    )
    parser.add_argument(
        "--team",
        help="Slug of team to which test results belong",
    )
    parser.add_argument(
        "--workflow",
        help="Name of the workflow to run",
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--testcase",
        "--testcases",
        dest="testcases",
        action="extend",
        nargs="+",
        help="One or more testcases to feed to the workflow",
    )
    group.add_argument(
        "--testcase-file",
        dest="testcase-file",
        help="Single file listing testcases to feed to the workflows",
    )

    parser.add_argument(
        "--config-file",
        help="Path to a configuration file",
        dest="file",
    )
    parser.add_argument(
        "--output-directory",
        help="Path to a local directory to store result files",
        dest="output-directory",
    )

    parser.add_argument(
        "--log-level",
        dest="log-level",
        choices=["debug", "info", "warn"],
        default="info",
        help="Level of detail with which events are logged",
    )
    parser.add_argument(
        "--save-as-binary",
        dest="save-as-binary",
        const=True,
        default=False,
        nargs="?",
        help="Save a copy of test results on local filesystem in binary format",
    )
    parser.add_argument(
        "--save-as-json",
        dest="save-as-json",
        const=True,
        default=False,
        nargs="?",
        help="Save a copy of test results on local filesystem in JSON format",
    )
    parser.add_argument(
        "--offline",
        const=True,
        default=False,
        nargs="?",
        help="Disables all communications with the Touca server",
    )
    parser.add_argument(
        "--overwrite",
        const=True,
        default=False,
        nargs="?",
        help="Overwrite result directory for testcase if it already exists",
    )
    parser.add_argument(
        "--colored-output",
        dest="colored-output",
        const=True,
        default=True,
        nargs="?",
        help="Use color in standard output",
    )


def _parse_cli_options(args) -> Dict[str, Any]:
    parser = ArgumentParser(
        description="Touca Regression Test",
        epilog="Visit https://touca.io/docs for more information",
    )
    prepare_parser(parser)
    parsed = vars(parser.parse_known_args(args)[0]).items()
    parsed = dict(filter(lambda x: x[1] is not None, parsed))
    # fix options with boolean values
    for k in [
        "save-as-binary",
        "save-as-json",
        "offline",
        "overwrite",
        "colored-output",
    ]:
        parsed[k] = True if parsed.get(k) in [True, "True", "true"] else False

    return parsed


class _ToucaErrorCode(IntEnum):
    MissingWorkflow = 1
    MissingSlugs = 2
    NoCaseMissingRemote = 3
    NoCaseEmptyRemote = 4


class _ToucaError(Exception):

    _errors: Dict[int, str] = {
        _ToucaErrorCode.MissingWorkflow: """\
            No workflow is registered.
            """,
        _ToucaErrorCode.MissingSlugs: """\
            Options {} are required when using this test framework.
            """,
        _ToucaErrorCode.NoCaseMissingRemote: """\
            Cannot proceed without a test case.
            Either use '--testcase' or '--testcase-file' to pass test cases
            or use '--api-key' and '--api-url' to let the library query
            the Touca Server to obtain and reuse the list of test cases
            submitted to the baseline version of this suite.
            """,
        _ToucaErrorCode.NoCaseEmptyRemote: """\
            Cannot proceed without a test case.
            Neither '--testcase' nor '--testcase-file' were provided.
            Attempted to query the Touca Server to obtain and reuse the
            list of test cases submitted to the baseline version of this
            suite but this suite has no previous version.
            """,
    }
    _wrapper = textwrap.TextWrapper(break_on_hyphens=False, width=80)

    def __init__(self, code: _ToucaErrorCode, *fmt_args: List[Any]):
        self._code = code
        if code in _ToucaError._errors:
            text = _ToucaError._errors.get(code).format(*fmt_args)
            wrapped = _ToucaError._wrapper.fill(textwrap.dedent(text))
            self._message = wrapped
        else:
            self._message = "Unknown Error"
        super().__init__(self._message)

    def __str__(self):
        return self._message


def _update_testcase_list(options: dict):
    """
    Use provided config options to find the final list of test cases to use
    for running the workflows. The following implementation assumes options
    `--testcases` and `--testcase-file` are mutually exclusive.
    """
    if options.get("testcases"):
        return
    if "testcase-file" in options:
        with open(options["testcase-file"], "rt") as file:
            keep = lambda x: x and not x.startswith("#")

            entries = list(filter(keep, file.read().splitlines()))
            options["testcases"] = entries
            return
    if options.get("offline"):
        elements = ["output-directory", "suite", "version"]
        version_dir = os.path.join(*map(options.get, elements))
        if not os.path.exists(version_dir):
            raise _ToucaError(_ToucaErrorCode.NoCaseMissingRemote)
        entries = os.listdir(version_dir)
        if not entries:
            raise _ToucaError(_ToucaErrorCode.NoCaseMissingRemote)
        options["testcases"] = entries
        return
    if any(k not in options for k in ["api-key", "api-url"]):
        raise _ToucaError(_ToucaErrorCode.NoCaseMissingRemote)
    options["testcases"] = Client.instance().get_testcases()
    if not options.get("testcases"):
        raise _ToucaError(_ToucaErrorCode.NoCaseEmptyRemote)


def _initialize(options: dict):
    from touca._options import find_home_path, update_options

    # Let the lower-level library consolidate the provided config options
    # including applying environment variables and processing long-format
    # api_url.
    update_options(options, options)

    # Check that team and suite are provided.
    missing = [k for k in ["team", "suite", "version"] if k not in options]
    if missing:
        raise _ToucaError(_ToucaErrorCode.MissingSlugs, ", ".join(missing))

    # Create directory to write logs and test results into
    options["output-directory"] = options.get(
        "output-directory", os.path.join(find_home_path(), "results")
    )
    os.makedirs(options.get("output-directory"), exist_ok=True)

    # Configure the lower-level Touca library
    if not Client.instance().configure(**options):
        raise RuntimeError(Client.instance().configuration_error())

    _update_testcase_list(options)


def _skip(options: dict, testcase: str):
    elements = ["output-directory", "suite", "version"]
    casedir = os.path.join(*map(options.get, elements), testcase)
    if options.get("save-as-binary"):
        return os.path.exists(os.path.join(casedir, "touca.bin"))
    if options.get("save-as-json"):
        return os.path.exists(os.path.join(casedir, "touca.json"))
    return False


class _Statistics:
    def __init__(self):
        from collections import defaultdict

        self._v = defaultdict(int)

    def inc(self, name: str):
        self._v[name] += 1

    def count(self, name: str):
        return self._v.get(name, 0)


class _Timer:
    def __init__(self):
        self._tics = {}
        self._times: Dict[str, timedelta] = {}

    def tic(self, name: str):
        self._tics[name] = datetime.utcnow()

    def toc(self, name: str):
        self._times[name] = datetime.utcnow() - self._tics.get(name)

    def count(self, name: str):
        return int(self._times.get(name).microseconds / 1e3)


class Workflow:
    """
    Base class meant to be used as a decorator.

    Registers the decorated function as a regression test workflow to be
    executed, once, for each test case.

    The following example demonstrates how this class should be used::

        @touca.Workflow
        def test_students(testcase: str):
            student = find_student(testcase)
            touca.assume("username", student.username)
            touca.check("fullname", student.fullname)
            touca.check("birth_date", student.dob)
            touca.check("gpa", calculate_gpa(student.courses))
    """

    def __init__(self, func):
        from functools import update_wrapper

        update_wrapper(self, func)
        self.__func = func
        if not hasattr(Workflow, "_workflows"):
            Workflow._workflows = []
        Workflow._workflows.append((func.__name__, self))

    def __call__(self, testcase: str):
        return self.__func(testcase)


def _filter_selected_workflow(options, workflows):
    if "workflow" not in options:
        return workflows
    filtered = [(k, v) for k, v in workflows if options["workflow"] == k]
    if not filtered:
        raise Exception("workflow: " + options["workflow"] + " does not exist")
    return filtered


def _run_workflow(options, workflow):
    offline = options.get("offline") or any(
        k not in options for k in ["api-key", "api-url"]
    )
    timer = _Timer()
    stats = _Statistics()
    printer = Printer(options)
    printer.print_header()
    timer.tic("__workflow__")

    for idx, testcase in enumerate(options.get("testcases")):
        elements = ["output-directory", "suite", "version"]
        casedir = os.path.join(*map(options.get, elements), testcase)

        if not options.get("overwrite") and _skip(options, testcase):
            printer.print_progress(timer, testcase, idx, "skip")
            stats.inc("skip")
            continue

        if os.path.exists(casedir):
            shutil.rmtree(casedir)
            os.makedirs(casedir)

        Client.instance().declare_testcase(testcase)
        timer.tic(testcase)

        errors = []
        try:
            workflow.__call__(testcase)

        except BaseException as err:
            errors.append(": ".join([err.__class__.__name__, str(err)]))
        except:
            errors.append("Unknown Error")

        timer.toc(testcase)
        status = "pass" if not errors else "fail"
        stats.inc(status)

        if not errors and options.get("save-as-binary"):
            Client.instance().save_binary(
                os.path.join(casedir, "touca.bin"), [testcase]
            )

        if not errors and options.get("save-as-json"):
            Client.instance().save_json(os.path.join(casedir, "touca.json"), [testcase])

        if not errors and not offline:
            Client.instance().post()

        printer.print_progress(timer, testcase, idx, status, errors)

        Client.instance().forget_testcase(testcase)

    timer.toc("__workflow__")
    printer.print_footer(stats, timer)

    if not offline:
        Client.instance().seal()


def run_workflows(args, workflows):
    from copy import deepcopy

    workflows = _filter_selected_workflow(args, workflows)
    Printer.print_app_header()
    for name, workflow in workflows:
        options = deepcopy(args)
        options["suite"] = name
        try:
            _initialize(options)
            _run_workflow(options, workflow)
        except RuntimeError as error:
            Printer.print_warning("Error when running workflow {}: {}", name, error)
    Printer.print_app_footer()


def run():
    """
    Runs registered workflows, one by one, for available the test cases.

    This function is intended to be called once from the main module as
    shown in the example below::

        if __name__ == "__main__":
            touca.run()

    :raises SystemExit:
        When configuration options specified as command line arguments,
        environment variables, or in a configuration file, have unexpected
        values or are in conflict with each other. Capturing this exception
        is not required.
    """
    if not hasattr(Workflow, "_workflows") or not Workflow._workflows:
        raise _ToucaError(_ToucaErrorCode.MissingWorkflow)
    try:
        cli_options = _parse_cli_options(sys.argv[1:])
        run_workflows(cli_options, Workflow._workflows)
    except _ToucaError as err:
        sys.exit(err)
    except Exception as err:
        sys.exit(f"Test failed: {err}")



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import sys
from pathlib import Path
from argparse import ArgumentParser
from distutils.version import LooseVersion
from loguru import logger
from touca._transport import Transport
from touca.cli._common import Operation


def _post(src_dir: Path, transport: Transport):
    src_dir = src_dir.with_name(src_dir.name + "-merged")
    if not src_dir.exists():
        logger.error(f"expected directory {src_dir} to exist")
        return False
    binaries = list(src_dir.rglob("**/*.bin"))
    if not binaries:
        logger.warning(f"{src_dir} has no result files")
        return False
    logger.debug(f"posting files in {src_dir}")
    for binary in binaries:
        logger.debug(f"posting {binary}")
        content = binary.read_bytes()
        transport._send_request(
            method="POST",
            path=f"/client/submit",
            body=content,
            content_type="application/octet-stream",
        )
        logger.debug(f"posted {binary}")
    logger.info(f"posted {src_dir}")
    return True


class Post(Operation):
    name = "post"
    help = "Submit binary archive files to remote server"

    def __init__(self, options: dict):
        self.__options = options

    @classmethod
    def parser(self, parser: ArgumentParser):
        parser.add_argument("src", help="path to directory with binary files")
        parser.add_argument("--api-key", help="Touca API Key", dest="api-key")
        parser.add_argument("--api-url", help="Touca API URL", dest="api-url")

    def run(self):
        from touca._options import update_options

        options = {
            k: self.__options.get(k)
            for k in ["api-key", "api-url"]
            if self.__options.get(k) is not None
        }
        options.update({"suite": "", "team": "", "version": ""})
        try:
            update_options(options, options)
        except ValueError as err:
            print(err, file=sys.stderr)
            return False

        api_key = options.get("api-key")
        api_url = options.get("api-url")

        src_dir = Path(self.__options.get("src")).expanduser().resolve()
        if not src_dir.exists():
            logger.error(f"directory {src_dir} does not exist")
            return False

        batchNames = []
        for batch_dir in src_dir.glob("*"):
            if not batch_dir.is_dir():
                continue
            if not batch_dir.name.endswith("-merged"):
                continue
            batchNames.append(batch_dir.name[:-7])

        if not batchNames:
            logger.info(f"found no valid result directory to post")
            return True

        # sort list of versions lexicographically
        batchNames.sort(key=LooseVersion)

        transport = Transport({"api-key": api_key, "api-url": api_url})
        try:
            transport.authenticate()
        except ValueError as err:
            print(err, file=sys.stderr)
            return False

        logger.info(f"posting batches one by one")
        for batchName in batchNames:
            batchDir = src_dir.joinpath(batchName)
            logger.info(f"posting {batchDir}")
            if not _post(batchDir, transport):
                logger.error(f"failed to post {batchDir}")
                return False
        logger.info("posted all result directories")
        return True



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import json
import os
import shutil
import subprocess
import tempfile
from argparse import ArgumentParser, ArgumentTypeError

import requests
from jsonschema import Draft3Validator
from loguru import logger
from touca.cli._common import Operation


def merge_dict(source: dict, destination: dict):
    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            merge_dict(value, node)
        else:
            destination[key] = value


def make_absolute_path(path: str, base: str = None) -> str:
    if os.path.isabs(path):
        return path
    if not base:
        base = os.path.dirname(os.path.realpath(__file__))
    if os.path.isfile(base):
        base = os.path.dirname(base)
    return os.path.abspath(os.path.join(base, path))


def parse_json_file(path: str) -> dict:
    logger.debug("parsing json file: {}", path)

    # check that the file exists
    if not os.path.exists(path):
        logger.warning("failed to find file: {}", path)
        return False

    # load file into memory, validate its format and parse its content
    try:
        with open(path, "rt") as file:
            data = file.read()
        return json.loads(data)
    except OSError as err:
        logger.warning("failed to read file: {}: {}", path, err)
    except ValueError as err:
        logger.warning("failed to parse file: {}: {}", path, err)

    return {}


def profile_parse(profile_path: str) -> dict:
    logger.debug("parsing profile: {}", profile_path)

    # parse profile
    config = parse_json_file(profile_path)
    if not config:
        logger.warning("failed to parse profile: {}", profile_path)
        return {}

    # if file extends a template, parse that template
    if "extends" in config:
        parent_file = make_absolute_path(config["extends"], profile_path)
        parent_config = profile_parse(parent_file)
        if not parent_config:
            logger.debug("failed to parse template: {}", parent_file)
            return {}
        del config["extends"]
        merge_dict(parent_config, config)

    return config


def profile_validate(config: dict) -> list:
    schema_path = make_absolute_path("./config/profile.schema.json")

    schema = parse_json_file(schema_path)
    if not schema:
        logger.warning("failed to profile schema: {}", schema_path)
        return False

    validator = Draft3Validator(schema)
    errors = sorted(validator.iter_errors(config, schema), key=str)
    for error in errors:
        logger.warning(
            "parameter {} in profile has unexpected value: {}",
            ".".join(list(error.relative_path)),
            error.message,
        )
    return not errors


def find_artifact_version(config: dict) -> str:
    cfg = config["artifactory"]
    fmt = "{base_url}/api/search/latestVersion?g={group}&a={name}&repos={repo}"
    query_url = fmt.format(
        base_url=cfg["base-url"], group=cfg["group"], name=cfg["name"], repo=cfg["repo"]
    )
    if "version-filter" in cfg:
        query_url += "&v=" + cfg["version-filter"]
    logger.debug("finding latest version: {}", query_url)
    artifact_version = requests.get(query_url).text
    logger.info("choosing version {}", artifact_version)
    return artifact_version


def build_artifact_download_url(config: dict, version: str) -> str:
    cfg = config["artifactory"]
    path = cfg["installer-msi-url"].format(
        repo=cfg["repo"], group=cfg["group"], name=cfg["name"], version=version
    )
    return f"{cfg['base-url']}/{path}"


def download_artifact(config: dict, tmpdir, artifact_version) -> str:
    download_url = build_artifact_download_url(config, artifact_version)
    msi_path = os.path.join(tmpdir, download_url.split("/")[-1])
    with open(msi_path, "wb") as tmpfile:
        logger.info("downloading artifact: {}", artifact_version)
        logger.debug("downloading: {}", download_url)
        with requests.get(download_url, stream=True) as response:
            response.raw.decode_content = True
            shutil.copyfileobj(response.raw, tmpfile)
            return msi_path


def install_artifact(config: dict, msi_path: str) -> bool:
    install_location = config["artifactory"]["installer-msi-location"]
    target_option = f'TARGET_DIR="{install_location}"'
    cmd = " ".join(["msiexec", "/i", msi_path, "/passive", "/qn", target_option])
    logger.info("installing artifact: {}", install_location)
    logger.debug("installing: {}", cmd)
    try:
        subprocess.check_output(cmd)
    except subprocess.CalledProcessError as err:
        logger.error(err)
        return False
    return True


def run_test(config: dict, artifact_version: str):
    install_location = config["artifactory"]["installer-msi-location"]
    cfg = config["execution"]
    test_config_path = os.path.join(install_location, cfg["config"])
    test_executable = os.path.join(install_location, cfg["executable"])
    cmd = [
        test_executable,
        "-c",
        test_config_path,
        "-r",
        artifact_version,
        "--suite",
        cfg["suite"],
    ]
    subprocess.run(cmd)
    return True


def archive_results(config: dict, artifact_version: str):
    install_location = config["artifactory"]["installer-msi-location"]
    cfg = config["execution"]
    test_config_path = os.path.join(install_location, cfg["config"])
    test_config = parse_json_file(test_config_path)
    output_dir = test_config["framework"]["output-dir"]
    src_dir = os.path.join(output_dir, cfg["suite"], artifact_version)
    dst_file = os.path.join(cfg["archive-dir"], cfg["suite"], artifact_version) + ".7z"
    cmd = ["C:\\Program Files\\7-Zip\\7z.exe", "a", dst_file, src_dir + "\\*"]
    subprocess.run(cmd)
    shutil.rmtree(src_dir)
    return True


class Run(Operation):
    name = "run"
    help = "Run regression test on a dedicated test server"

    def __init__(self, options: dict):
        self.__options = options

    @classmethod
    def parser(self, parser: ArgumentParser):
        parser.add_argument(
            "-p",
            "--profile",
            type=str,
            help="path to the profile file",
            action="store",
            required=True,
            default=None,
        )
        parser.add_argument(
            "-r",
            "--revision",
            type=str,
            help="specific version to run",
            action="store",
            required=False,
            default=None,
        )
        return parser

    def run(self):
        profile_path = self.__options.get("profile")

        # parse profile_name

        profile_name = os.path.splitext(
            os.path.basename(os.path.abspath(profile_path))
        )[0]
        logger.debug("running profile: {}", profile_name)

        # check that profile is a valid json file

        config = profile_parse(os.path.abspath(profile_path))
        if not config:
            logger.error("failed to parse profile: {}", profile_path)
            return False

        # validate configuration parameters

        if not profile_validate(config):
            logger.error("failed to validate test profile")
            return False

        # find version of the test artifact

        artifact_version = (
            find_artifact_version(config)
            if "revision" in self.__options
            else self.__options.get("revision")
        )
        if not artifact_version:
            logger.error("failed to find artifact version: {}", artifact_version)
            return False

        # check if version is already executed

        archive_root = make_absolute_path(
            config["execution"]["archive-dir"], profile_path
        )
        if os.path.exists(
            os.path.join(archive_root, profile_name, artifact_version) + ".7z"
        ):
            logger.info("{}/{} is already executed", artifact_version, profile_name)
            return True

        # download and install the test artifact

        with tempfile.TemporaryDirectory(prefix="touca_runner_artifact") as tmpdir:
            msi_path = download_artifact(config, tmpdir, artifact_version)
            if not msi_path:
                logger.error("failed to download artifact: {}", artifact_version)
                return False
            if not install_artifact(config, msi_path):
                logger.error("failed to install artifact: {}", artifact_version)
                return False

        # run the test on a separate thread

        if not run_test(config, artifact_version):
            logger.error("failed to run the test")
            return False

        # archive the test results

        if not archive_results(config, artifact_version):
            logger.error("failed to archive test results")
            return False

        logger.info(
            "test complete for {}/{}", config["execution"]["suite"], artifact_version
        )
        return True



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

"""Touca SDK for Python."""

import os
from setuptools import setup


def get_file_content(file_name):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, file_name), "rb") as file:
        return file.read().decode("utf-8")


def get_version():
    for line in get_file_content("touca/__init__.py").splitlines():
        if line.startswith("__version__"):
            delimeter = '"' if '"' in line else "'"
            return line.split(delimeter)[1]
    raise RuntimeError("Unable to find version string.")


version = get_version()
repo_url = "https://github.com/trytouca/trytouca/tree/main/sdk/python"

setup(
    name="touca",
    version=version,
    license="Apache-2.0",
    description="Touca SDK for Python",
    packages=["touca"],
    package_data={
        "touca": [
            "py.typed",
            "cli/config/profile.schema.json",
            "cli/*.py",
            "plugins/*.py",
        ]
    },
    keywords=["touca", "snapshot testing", "regression testing"],
    long_description=get_file_content("README.md"),
    long_description_content_type="text/markdown",
    author="Touca, Inc.",
    author_email="support@touca.io",
    url=repo_url,
    download_url="{}/archive/{}.tar.gz".format(repo_url, version),
    project_urls={
        "Changelog": "{}/blob/master/Changelog.md".format(repo_url),
        "Documentation": "https://touca.io/docs",
        "Source": repo_url,
        "Twitter": "https://twitter.com/trytouca",
    },
    entry_points={"console_scripts": ["touca=touca.cli.__main__:main"]},
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    install_requires=[
        "urllib3>=1.26.9",
        "certifi>=2021.5.30",
        "flatbuffers>=2.0",
        # The core library has no dependency on dataclasses at the moment.
        # This may change in the future. For now, we are specifying this
        # dependency to allow users to run getting started examples that
        # may be using dataclasses.
        'dataclasses; python_version<"3.7"',
        "colorama>=0.4.4",
        "jsonschema",
        "loguru>=0.6.0",
        "packaging>=21.3",
        "py7zr>=0.18.5",
        "requests>=2.27.1",
    ],
    extras_require={
        "dev": [
            "black>=21.6b0",
            "pytest>=6.2.4",
            "pytest-cov>=2.12.1",
            "Sphinx>=4.0.2",
            "sphinx-rtd-theme>=0.5.2",
            "tox>=3.9.0",
        ]
    },
)



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

"""Touca plugin for Django test framework."""

from django.test.runner import DiscoverRunner
from unittest import TestSuite, TestCase, TextTestRunner
import touca


class _TestCase(TestCase):
    """Specialized TestCase to be handled by the Touca runner for Django."""

    def __init__(self, case: TestCase):
        """Clone a given ``TestCase`` instance for added functionality."""
        super().__init__()
        self.case = case

    def runTest(self):
        """
        Execute test case.

        Includes a simplified version of the default logic in the Touca test
        framework to be run for each test case.
        """
        touca.declare_testcase(self.case._testMethodName)
        self.case.run()
        instance = touca._client.Client.instance()
        active_case = instance._cases.get(instance._active_testcase_name())
        if instance._transport and (
            len(active_case._results) or len(active_case._tics)
        ):
            touca.post()
        if instance.is_configured():
            touca.forget_testcase(self.case._testMethodName)


class _TestRunner(TextTestRunner):
    """Specialized ``TextTestRunner`` to be used by the Touca runner for Django."""

    def run(self, test: TestSuite):
        """Run a given test suite."""
        test._tests = [_TestCase(x) for x in test._tests]
        return super().run(test)


class Runner(DiscoverRunner):
    """Custom test runner to use with Django test framework."""

    def __init__(
        self, api_key=None, api_url=None, revision=None, offline=None, **kwargs
    ):
        """
        Create a test runner instance derived from Django's ``DiscoverRunner``.

        To be passed to the Django test framework via the ``--testrunner``
        command line option.
        """
        from os import environ

        super().__init__(**kwargs)
        self.touca_options = {
            "api-key": environ.get("TOUCA_API_KEY", api_key),
            "api-url": environ.get("TOUCA_API_URL", api_url),
            "version": environ.get("TOUCA_TEST_VERSION", revision),
            "offline": True if offline in [True, "True", "true"] else False,
        }
        self.touca_options = {
            k: v for k, v in self.touca_options.items() if v is not None
        }

    def setup_test_environment(self, **kwargs):
        """Configure the Touca client after performing default setup."""
        super(Runner, self).setup_test_environment(**kwargs)
        touca.configure(**self.touca_options)

    def run_tests(self, test_labels, extra_tests=None, **kwargs):
        """Run Django test cases with extra Touca functionalities."""
        self.test_runner = _TestRunner
        super().run_tests(test_labels, extra_tests, **kwargs)

    @classmethod
    def add_arguments(cls, parser):
        """
        Add common Touca command-line options to Django test framework.

        Extends the set of command-line options supported by the Django test
        framework to allow for changing Touca behavior from the command-line.
        """
        DiscoverRunner.add_arguments(parser)
        parser.add_argument("--api-key", dest="api-key", help="Touca API Key")
        parser.add_argument("--api-url", dest="api-url", help="Touca API URL")
        parser.add_argument("--revision", dest="revision", help="Touca Test Version")
        parser.add_argument(
            "--offline",
            action="store",
            default=False,
            help="Disables all communications with the Touca server",
        )



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

"""
Entry-point to the Touca SDK for Python.

You can install this sdk via ``pip install touca`` and import it in your code via::

    import touca

Alternatively, you can import individual functions which may be useful in rare
cases if and when you want to call them from production code::

    from touca import check

If you are just getting started with Touca, we generally recommend that you
install the SDK as a development-only dependency.
"""

from typing import Any, Callable, Dict, List, Type
from touca._client import Client
from touca._runner import Workflow, run
from touca._utils import scoped_timer

__version__ = "1.5.4"


def clientmethod(f):
    import inspect

    f.__doc__ = inspect.getdoc(getattr(Client, f.__name__))
    return f


@clientmethod
def configure(**kwargs) -> bool:
    return Client.instance().configure(**kwargs)


@clientmethod
def is_configured() -> bool:
    return Client.instance().is_configured()


@clientmethod
def configuration_error() -> str:
    return Client.instance().configuration_error()


@clientmethod
def get_testcases() -> List[str]:
    return Client.instance().get_testcases()


@clientmethod
def declare_testcase(name: str):
    Client.instance().declare_testcase(name)


@clientmethod
def forget_testcase(name: str):
    Client.instance().forget_testcase(name)


@clientmethod
def check(key: str, value: Any):
    Client.instance().check(key, value)


@clientmethod
def assume(key: str, value: Any):
    Client.instance().assume(key, value)


@clientmethod
def add_array_element(key: str, value: Any):
    Client.instance().add_array_element(key, value)


@clientmethod
def add_hit_count(key: str):
    Client.instance().add_hit_count(key)


@clientmethod
def add_metric(key: str, milliseconds: int):
    Client.instance().add_metric(key, milliseconds)


@clientmethod
def start_timer(key: str):
    Client.instance().start_timer(key)


@clientmethod
def stop_timer(key: str):
    Client.instance().stop_timer(key)


@clientmethod
def add_serializer(datatype: Type, serializer: Callable[[Any], Dict]):
    Client.instance().add_serializer(datatype, serializer)


@clientmethod
def save_binary(key: str, cases: List[str] = []):
    Client.instance().save_binary(key, cases)


@clientmethod
def save_json(key: str, cases: List[str] = []):
    Client.instance().save_json(key, cases)


@clientmethod
def post():
    Client.instance().post()


@clientmethod
def seal():
    Client.instance().seal()



# Copyright 2021 Touca, Inc. Subject to Apache-2.0 License.

from typing import Any, Callable, Dict, List, ValuesView, Type
from threading import get_ident
from touca._case import Case


def casemethod(func):
    """ """
    import inspect
    from functools import wraps

    func.__doc__ = inspect.getdoc(getattr(Case, func.__name__))

    @wraps(func)
    def wrapper(self, *args):
        element = self._active_testcase_name()
        if not element:
            return
        testcase = self._cases.get(element)
        method = getattr(testcase, func.__name__)
        retval = func(self, *args)
        if not retval:
            method(*args)
        else:
            method(args[0], retval)

    return wrapper


class Client:
    """ """

    @classmethod
    def instance(cls):
        if not hasattr(cls, "_instance"):
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        from touca._types import TypeHandler

        self._cases: Dict[str, Case] = dict()
        self._configured = False
        self._configuration_error = str()
        self._options = dict()
        self._threads_case = str()
        self._threads_cases: Dict[int, str] = dict()
        self._transport = None
        self._type_handler = TypeHandler()

    def _active_testcase_name(self) -> str:
        if not self._configured:
            return None
        if self._options.get("concurrency"):
            return self._threads_case
        return self._threads_cases.get(get_ident())

    def _make_transport(self) -> bool:
        from touca._transport import Transport

        keys = ["api-key", "api-url", "team", "suite", "version"]
        if self._options.get("offline") is True:
            return False
        if not all(k in self._options for k in keys):
            return False
        if not self._transport:
            self._transport = Transport(self._options)
            return True
        is_different = lambda k: self._transport._options.get(k) != self._options.get(k)
        if list(filter(is_different, self._options.keys())):
            self._transport.update_options(self._options)
        return True

    def _serialize(self, cases: ValuesView[Case]) -> bytearray:
        from sys import version_info
        from flatbuffers import Builder
        import touca._schema as schema

        builder = Builder(1024)

        message_buffers = []
        for item in reversed(list(cases) if version_info < (3, 8) else cases):
            buffer = builder.CreateByteVector(item.serialize())
            schema.MessageBufferStart(builder)
            schema.MessageBufferAddBuf(builder, buffer)
            message_buffers.append(schema.MessageBufferEnd(builder))

        schema.MessageBufferStartBufVector(builder, len(message_buffers))
        for msg_buf in reversed(message_buffers):
            builder.PrependUOffsetTRelative(msg_buf)
        fbs_msg_buffers = builder.EndVector()

        schema.MessagesStart(builder)
        schema.MessagesAddMessages(builder, fbs_msg_buffers)
        fbs_messages = schema.MessagesEnd(builder)
        builder.Finish(fbs_messages)

        return builder.Output()

    def _prepare_save(self, path, cases):
        import os

        if os.path.dirname(path):
            os.makedirs(os.path.dirname(path), exist_ok=True)
        if cases:
            return [self._cases[x] for x in self._cases if x in cases]
        return self._cases.values()

    def configure(self, **kwargs) -> bool:
        """
        Configures the Touca client. Must be called before declaring test cases
        and adding results to the client. Should be regarded as a potentially
        expensive operation. Should be called only from your test environment.

        :py:meth:`~configure` takes a variety of configuration parameters
        documented below. All of these parameters are optional. Calling this
        function without any parameters is possible: the client can capture
        behavior and performance data and store them on a local filesystem
        but it will not be able to post them to the Touca server.

        In most cases, You will need to pass API Key and API URL during the
        configuration. The code below shows the common pattern in which API URL
        is given in long format (it includes the team slug and the suite slug)
        and API Key as well as the version of the code under test are specified
        as environment variables ``TOUCA_API_KEY`` and ``TOUCA_TEST_VERSION``,
        respectively::

            touca.configure(api_url='https://api.touca.io/@/acme/students')

        As long as the API Key and API URL to the Touca server are known to
        the client, it attempts to perform a handshake with the Touca Server to
        authenticate with the server and obtain the list of known test cases
        for the baseline version of the specified suite. You can explicitly
        disable this handshake in rare cases where you want to prevent ever
        communicating with the Touca server.

        You can call :py:meth:`~configure` any number of times. The client
        preserves the configuration parameters specified in previous calls to
        this function.

        :type file: str, optional
        :param file:
            (optional) Path to a configuration file in JSON format with a
            top-level "touca" field that may list any number of configuration
            parameters for this function. When used alongside other parameters,
            those parameters would override values specified in the file.

        :type api_key: str, optional
        :param api_key:
            (optional) API Key issued by the Touca server that
            identifies who is submitting the data. Since the value should be
            treated as a secret, we recommend that you pass it as an environment
            variable ``TOUCA_API_KEY`` instead.

        :type api_url: str, optional
        :param api_url:
            (optional) URL to the Touca server API.
            Can be provided either in long format like
            ``https://api.touca.io/@/myteam/mysuite/version``
            or in short format like ``https://api.touca.io``.
            If the team, suite, or version are specified, you do not need to
            specify them separately.

        :type team: str, optional
        :param team:
            (optional) slug of your team on the Touca server.

        :type suite: str, optional
        :param suite:
            slug of the suite on the Touca server that corresponds to your
            workflow under test.

        :type version: str, optional
        :param version:
            version of your workflow under test.

        :type offline: bool, optional
        :param offline:
            determines whether client should connect with the Touca server
            during the configuration. Defaults to ``False`` when ``api_url``
            or ``api_key`` are provided.

        :type concurrency: bool, optional
        :param concurrency:
            determines whether the scope of test case declaration is bound to
            the thread performing the declaration, or covers all other threads.
            Defaults to ``True``.
            If set to ``True``, when a thread calls :py:meth:`~declare_testcase`,
            all other threads also have their most recent test case changed to
            the newly declared test case and any subsequent call to data
            capturing functions such as :py:meth:`~check`
            will affect the newly declared test case.

        :return: ``True`` if client is ready to capture data.
        :rtype: bool
        """
        from touca._options import update_options

        self._configuration_error = ""
        try:
            update_options(self._options, kwargs)
            if self._make_transport():
                self._transport.authenticate()
        except (RuntimeError, ValueError) as err:
            self._configuration_error = f"Configuration failed: {err}"
            return False
        self._configured = True
        return True

    def is_configured(self) -> bool:
        """
        Checks if previous call(s) to :py:meth:`~configure` have set the right
        combination of configuration parameters to enable the client to perform
        expected tasks.

        We recommend that you perform this check after client configuration and
        before calling other functions of the library::

            if not touca.is_configured():
                print(touca.configuration_error())
                sys.exit(1)

        At a minimum, the client is considered configured if it can capture
        test results and store them locally on the filesystem. A single call
        to :py:meth:`~configure` without any configuration parameters can help
        us get to this state. However, if a subsequent call to :py:meth:`~configure`
        sets the parameter ``api_url`` in short form without specifying
        parameters such as ``team``, ``suite`` and ``version``, the client
        configuration is incomplete: We infer that the user intends to submit
        results but the provided configuration parameters are not sufficient
        to perform this operation.

        :return: ``True`` if the client is properly configured
        :rtype: bool
        :see also: :py:meth:`~configure`
        """
        return self._configured

    def configuration_error(self) -> str:
        """
        Provides the most recent error, if any, that is encountered during
        client configuration.

        :return: short description of the most recent configuration error
        :rtype: str
        """
        return self._configuration_error

    def get_testcases(self) -> List[str]:
        """
        Queries the Touca server for the list of testcases that are submitted
        to the baseline version of this suite.

        :raises RuntimeError:
            when called on the client that is not configured to communicate
            with the Touca server.

        :return: list of test cases of the baseline version of this suite
        """
        if not self._transport:
            raise RuntimeError("client not configured to perform this operation")
        return self._transport.get_testcases()

    def declare_testcase(self, name: str):
        """
        Declares name of the test case to which all subsequent results will be
        submitted until a new test case is declared.

        If configuration parameter ``concurrency`` is set to ``"enabled"``, when
        a thread calls `declare_testcase` all other threads also have their most
        recent testcase changed to the newly declared one. Otherwise, each
        thread will submit to its own testcase.

        :param name:
            name of the testcase to be declared
        """
        if not self._configured:
            return
        if name not in self._cases:
            self._cases[name] = Case(
                team=self._options.get("team"),
                suite=self._options.get("suite"),
                version=self._options.get("version"),
                name=name,
            )
        self._threads_case = name
        self._threads_cases[get_ident()] = name
        return self._cases.get(name)

    def forget_testcase(self, name: str):
        """
        Removes all logged information associated with a given test case.

        This information is removed from memory, such that switching back to
        an already-declared or already-submitted test case would behave similar
        to when that test case was first declared. This information is removed,
        for all threads, regardless of the configuration option ``concurrency``.
        Information already submitted to the server will not be removed from
        the server.

        This operation is useful in long-running regression test frameworks,
        after submission of test case to the server, if memory consumed by
        the client library is a concern or if there is a risk that a future
        test case with a similar name may be executed.

        :param name:
            name of the testcase to be removed from memory

        :raises RuntimeError:
            when called with the name of a test case that was never declared
        """
        if name not in self._cases:
            raise RuntimeError(f'test case "{name}" was never declared')
        del self._cases[name]

    @casemethod
    def check(self, key: str, value: Any):
        return self._type_handler.transform(value)

    @casemethod
    def assume(self, key: str, value: Any):
        return self._type_handler.transform(value)

    @casemethod
    def add_array_element(self, key: str, value: Any):
        return self._type_handler.transform(value)

    @casemethod
    def add_hit_count(self, key: str):
        return

    @casemethod
    def add_metric(self, key: str, milliseconds: int):
        return milliseconds

    @casemethod
    def start_timer(self, key: str):
        return

    @casemethod
    def stop_timer(self, key: str):
        return

    def add_serializer(self, datatype: Type, serializer: Callable[[Any], Dict]):
        """
        Registers custom serialization logic for a given custom data type.

        Calling this function is rarely needed. The library already handles
        all custom data types by serializing all their properties. Custom
        serializers allow you to exclude a subset of an object properties
        during serialization.

        :param datattype: type to be serialized
        :param serializer: function that converts any object of the given type
            to a dictionary.
        """
        self._type_handler.add_serializer(datatype, serializer)

    def save_binary(self, path: str, cases: list):
        """
        Stores test results and performance benchmarks in binary format
        in a file of specified path.

        Touca binary files can be submitted at a later time to the Touca
        server.

        We do not recommend as a general practice for regression test tools
        to locally store their test results. This feature may be helpful for
        special cases such as when regression test tools have to be run in
        environments that have no access to the Touca server (e.g. running
        with no network access).

        :param path: path to file in which test results and performance
            benchmarks should be stored
        :param cases: names of test cases  whose results should be stored.
            If a set is not specified or is set as empty, all test cases will
            be stored in the specified file.
        """
        items = self._prepare_save(path, cases)
        content = self._serialize(items)
        with open(path, mode="wb") as file:
            file.write(content)

    def save_json(self, path: str, cases: list):
        """
        Stores test results and performance benchmarks in JSON format
        in a file of specified path.

        This feature may be helpful during development of regression tests
        tools for quick inspection of the test results and performance metrics
        being captured.

        :param path: path to file in which test results and performance
            benchmarks should be stored
        :param cases: names of test cases  whose results should be stored.
            If a set is not specified or is set as empty, all test cases will
            be stored in the specified file.
        """
        from json import dumps

        items = self._prepare_save(path, cases)
        content = dumps([testcase.json() for testcase in items])
        with open(path, mode="wt") as file:
            file.write(content)

    def post(self):
        """
        Submits all test results recorded so far to Touca server.

        It is possible to call :py:meth:`~post` multiple times during runtime
        of the regression test tool. Test cases already submitted to the server
        whose test results have not changed, will not be resubmitted.
        It is also possible to add test results to a testcase after it is
        submitted to the server. Any subsequent call to :py:meth:`~post` will
        resubmit the modified test case.

        :raises RuntimeError:
            when called on the client that is not configured to communicate
            with the Touca server.
        """
        if not self._transport:
            raise RuntimeError("client not configured to perform this operation")
        if not self._transport._token:
            raise RuntimeError("client not authenticated")
        self._transport.post(content=self._serialize(self._cases.values()))

    def seal(self):
        """
        Notifies the Touca server that all test cases were executed for this
        version and no further test result is expected to be submitted.
        Expected to be called by the test tool once all test cases are executed
        and all test results are posted.

        Sealing the version is optional. The Touca server automatically
        performs this operation once a certain amount of time has passed since
        the last test case was submitted. This duration is configurable from
        the "Settings" tab in "Suite" Page.

        :raises RuntimeError:
            when called on the client that is not configured to communicate
            with the Touca server.
        """
        if not self._transport:
            raise RuntimeError("client not configured to perform this operation")
        if not self._transport._token:
            raise RuntimeError("client not authenticated")
        self._transport.seal()



import flatbuffers


def ArrayStart(builder):
    builder.StartObject(1)


def ArrayAddValues(builder, values):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(values), 0
    )


def ArrayStartValuesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def ArrayEnd(builder):
    return builder.EndObject()


def BoolStart(builder):
    builder.StartObject(1)


def BoolAddValue(builder, value):
    builder.PrependBoolSlot(0, value, 0)


def BoolEnd(builder):
    return builder.EndObject()


def DoubleStart(builder):
    builder.StartObject(1)


def DoubleAddValue(builder, value):
    builder.PrependFloat64Slot(0, value, 0.0)


def DoubleEnd(builder):
    return builder.EndObject()


def IntStart(builder):
    builder.StartObject(1)


def IntAddValue(builder, value):
    builder.PrependInt64Slot(0, value, 0)


def IntEnd(builder):
    return builder.EndObject()


def MessageStart(builder):
    builder.StartObject(4)


def MessageAddMetadata(builder, metadata):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(metadata), 0
    )


def MessageAddResults(builder, results):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(results), 0
    )


def MessageAddMetrics(builder, metrics):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(metrics), 0
    )


def MessageEnd(builder):
    return builder.EndObject()


def MessageBufferStart(builder):
    builder.StartObject(1)


def MessageBufferAddBuf(builder, buf):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(buf), 0
    )


def MessageBufferStartBufVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def MessageBufferEnd(builder):
    return builder.EndObject()


def MessagesStart(builder):
    builder.StartObject(1)


def MessagesAddMessages(builder, messages):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(messages), 0
    )


def MessagesStartMessagesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def MessagesEnd(builder):
    return builder.EndObject()


def MetadataStart(builder):
    builder.StartObject(6)


def MetadataAddTestsuite(builder, testsuite):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(testsuite), 0
    )


def MetadataAddVersion(builder, version):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(version), 0
    )


def MetadataAddTestcase(builder, testcase):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(testcase), 0
    )


def MetadataAddBuiltAt(builder, builtAt):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(builtAt), 0
    )


def MetadataAddTeamslug(builder, teamslug):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(teamslug), 0
    )


def MetadataEnd(builder):
    return builder.EndObject()


def MetricStart(builder):
    builder.StartObject(2)


def MetricAddKey(builder, key):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(key), 0
    )


def MetricAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def MetricEnd(builder):
    return builder.EndObject()


def MetricsStart(builder):
    builder.StartObject(1)


def MetricsAddEntries(builder, entries):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(entries), 0
    )


def MetricsStartEntriesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def MetricsEnd(builder):
    return builder.EndObject()


def ObjectStart(builder):
    builder.StartObject(2)


def ObjectAddKey(builder, key):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(key), 0
    )


def ObjectAddValues(builder, values):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(values), 0
    )


def ObjectStartValuesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def ObjectEnd(builder):
    return builder.EndObject()


def ObjectMemberStart(builder):
    builder.StartObject(2)


def ObjectMemberAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0
    )


def ObjectMemberAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def ObjectMemberEnd(builder):
    return builder.EndObject()


def ResultStart(builder):
    builder.StartObject(3)


def ResultAddKey(builder, key):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(key), 0
    )


def ResultAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def ResultAddTyp(builder, typ):
    builder.PrependUint8Slot(2, typ, 1)


def ResultEnd(builder):
    return builder.EndObject()


class ResultType(object):
    Check = 1
    Assert = 2


def ResultsStart(builder):
    builder.StartObject(1)


def ResultsAddEntries(builder, entries):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(entries), 0
    )


def ResultsStartEntriesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def ResultsEnd(builder):
    return builder.EndObject()


def StringStart(builder):
    builder.StartObject(1)


def StringAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def StringEnd(builder):
    return builder.EndObject()


class Type(object):
    NONE = 0
    Bool = 1
    Int = 2
    UInt = 3
    Float = 4
    Double = 5
    String = 6
    Object = 7
    Array = 8


def TypeWrapperStart(builder):
    builder.StartObject(2)


def TypeWrapperAddValueType(builder, valueType):
    builder.PrependUint8Slot(0, valueType, 0)


def TypeWrapperAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def TypeWrapperEnd(builder):
    return builder.EndObject()



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

from abc import ABC, abstractmethod


class Operation(ABC):
    @abstractmethod
    def run(self) -> bool:
        pass


def invalid_subcommand(cls):
    from argparse import ArgumentParser
    import sys

    parser = ArgumentParser(prog=f"touca {cls.name}", description=cls.help)
    cls.parser(parser)
    parser.print_help(file=sys.stderr)
    return False
